"""
This file is an example on how to set experiments for MRE. 
The file experiments does all the experiments and produce a very
similar array of numbers that appear in the table of the papers. 

For details on how this file works and on how to make your own
experiments, see experiments_template.py

Even before reading experiments_template.py (in order to understand
how the whole thing works (which isn't all that hard)), 
it is a good idea to play with the parameters. 
For example, try changing arith_no_mul to arith, to see how hard it is
for the system to understand multiplication from a higher order
definition. 
The family experiments
appear to be quite robust. The arithmetic problem seems to actually be
harder, especially the multiplication part. 


To try out MRE, type import p8.mre.experimets as e

e.run_experiment(e.EX1)

or 

e.run_experiment(e.EX2)



...


e.run_experiment(e.EX4)

See this file for what EX1 ... EX4 mean. 
You can also change them. 
Once you do run_experiment, it'll keep on running for a long time and
occasionally spit out numbers which are the results of the
simulation. 

You might change the basic parameters of the simulation by playing
with simulation_factory.py (the weight decay and the global scale of
the function is there (somehow, BFGS is sensitive to multiplying the
objective function (and the gradients) by 1e-3 : in my old experiments
it was necessary, but now it seems to be doing more fine.


"""

import p8.mre.utils as utils
from pylab import rand, ones, zeros
import p8.mre.mre as mre
reload (mre)
import p8.mre.simulation_factory as simulation_factory
reload (simulation_factory)


class Experiment:
    def __init__(self, 
                 name, 
                 data, 
                 use_meta, 
                 dim,

                 how_many_to_remove=None,
                 relation_to_remove=None):
        """
        The experiment class describes an experiment. It exists so
        that we won't need to make function calls with very many
        parameters; in stead, we'll just define a list of
        experiments, and all will be good.


        It contains:
        the data
        whether meta relations should be used (or not)
        the dimensionality of the experiment.

        And methods of creating the datasets:
        how_many_to_remove and relation_to_remove are lists of the
        same length.

        if how_many_to_remove[i] = h and h is not None, then we make
        sure that relation_to_remove[i] is none, and in the i'th
        experiment we'll create a random training set where h basic
        relations are removed at random.

        if relation_to_remove[i] = r and r is not None, hten we make
        sure that how_many_to_remove[i] is none, and in the i'th
        experiment we'll remove all instances of relation r from the
        set of training basic relation; this will let us test the
        usefulness of the higher order relations. 
        
        """

        self.name = name
        self.data = data
        self.use_meta = use_meta
        self.dim = dim

        if how_many_to_remove is None:
            how_many_to_remove = [None] * len(relation_to_remove)

        if relation_to_remove is None:
            relation_to_remove = [None] * len(how_many_to_remove)

        self.how_many_to_remove = how_many_to_remove
        self.relation_to_remove = relation_to_remove


        assert(len(how_many_to_remove) == len(relation_to_remove))

        num_exper = len(relation_to_remove)
        self.num_exper = num_exper

        ## create the list of datasets:
        self.train_O_R0_O_inds_list = [None] * num_exper
        self.train_O_R1_R0_inds_list = [None] * num_exper
        self.train_R0_R2_R0_inds_list = [None] * num_exper


        for i, (how_many, relation) in enumerate(zip(how_many_to_remove,
                                                     relation_to_remove)):

            assert(not (how_many is None and 
                        relation is None))

            assert(not (how_many is not None and 
                        relation is not None))
            
            if relation is not None:
                self.train_O_R0_O_inds_list[i] = (
                    utils.train_indicies_without_relation(data.O_R0_O,
                                                          relation))
            elif how_many is not None:
                self.train_O_R0_O_inds_list[i] = (
                    utils.random_train_indicies(len(data.O_R0_O),
                                                len(data.O_R0_O) - how_many))

            else: raise Exception  ("something is very weird here")


            if use_meta:
                self.train_O_R1_R0_inds_list[i] = (
                    ones(len(data.O_R1_R0)).astype(bool))

                self.train_R0_R2_R0_inds_list[i] = (
                    ones(len(data.R0_R2_R0)).astype(bool))

            else:
                self.train_O_R1_R0_inds_list[i] = (
                    zeros(len(data.O_R1_R0)).astype(bool))

                self.train_R0_R2_R0_inds_list[i] = (
                    zeros(len(data.R0_R2_R0)).astype(bool))





## we are now ready to define the experiments:
import p8.mre.data.family as family
import p8.mre.data.arith as arith
import p8.mre.data.arith_no_mul as arith_no_mul

EX1 = Experiment(name='Family (not higher order)',
                 data=family,
                 use_meta=False,
                 dim=4,

                 how_many_to_remove=[10,10,10,10,10, 
                                     20,20,20,20,20, 
                                     30,30,30,30,30])

EX2 = Experiment(name='Higher order family',
                 data=family,
                 use_meta=True,
                 dim=4,

                 relation_to_remove=[1, 1, 1, 1, 1,
                                     4, 4, 4, 4, 4,
                                     6, 6, 6, 6, 6,
                                     11,11,11,11,11])

EX3 = Experiment(name='Arithmetic (not higher order)',
                 data=arith,
                 use_meta=False,
                 dim=4,

                 how_many_to_remove=[30, 30, 30, 30, 30, 
                                     60, 60, 60, 60, 60,
                                     90, 90, 90, 90, 90])

EX4 = Experiment(name='Higher order arithmetic',
                 data=arith_no_mul, 
                 use_meta=True,
                 dim=4,

                 relation_to_remove=[1, 1, 1, 1, 1,
                                     4, 4, 4, 4, 4,
                                     6, 6, 6, 6, 6,
                                     10,10,10,10,10])
                                     
                                     
                                     


def run_experiment(exper):
    data = exper.data

    print ('%s: I will now run %s experiments.\n' % 
           (exper.name, exper.num_exper))

    for i in range(exper.num_exper):
        train_O_R0_O_inds = exper.train_O_R0_O_inds_list[i]
        train_O_R1_R0_inds = exper.train_O_R1_R0_inds_list[i]
        train_R0_R2_R0_inds = exper.train_R0_R2_R0_inds_list[i]

        assert(train_O_R0_O_inds is not None)
        assert(train_O_R1_R0_inds is not None)
        assert(train_R0_R2_R0_inds is not None)

        # Create an MRE object. See mre/mre.py for precise details of its
        # parameters. Briefly, it gets all the training sets, all the
        # training indicies, the numbers of each objects, relations, and
        # meta relations.
        mre_object = mre.Mre(
            exper.dim,
            
            # the numbers of different objects/relations
            data.NUM_O,
            data.NUM_R0,
            data.NUM_R1,
            data.NUM_R2,
                     
            # the datasets
            data.O_R0_O,
            data.O_R1_R0,
            data.R0_R2_R0,
                     
            # the training set
            train_O_R0_O_inds, 
            train_O_R1_R0_inds,
            train_R0_R2_R0_inds)

        relation_to_remove = exper.relation_to_remove[i]
        how_many_to_remove = exper.how_many_to_remove[i]

        print ('%s: Running experiment %s:.\n\
relation_to_remove = %s.\n\
how_many_to_remove = %s.\n\
Train O_R0_O (basic relations) size = %s / %s (total).\n\
Train O_R1_R0 meta rel size = %s / %s (total).\n\
Train R0_R2_R0 meta rel size = %s / %s (total).\n\
It will take a few minutes (maybe even 5, but not much more (on an 3.5GHz machine)).' 
               % (exper.name, i, 
                  relation_to_remove,
                  how_many_to_remove,
                  sum(train_O_R0_O_inds), len(data.O_R0_O),
                  sum(train_O_R1_R0_inds), len(data.O_R1_R0),
                  sum(train_R0_R2_R0_inds), len(data.R0_R2_R0)))

        learned_mre = simulation_factory.run(mre_object, 
                                             check_grad=False)


        train_errors, test_errors = learned_mre.nearest_neighbor()


        print '%s: Train errors = %s / %s' % (exper.name, 
                                              train_errors[0],
                                              train_errors[1])

        print '%s: Test errors = %s / %s' % (exper.name, 
                                             test_errors[0],
                                             test_errors[1])
        print '\n***\n\n'




