"""
This file implements the functions that compute the gradients with
respect to MRE's cost function. 
IT implements a function that computes the performance of the

algorithm that answers queries with objects that are close to RA.
(see paper) on the test and the training set.

It also implements a class that
handles the various datasets that MRE has to deal with:
(for reference, they are O_R0_O, O_R1_R0, R0_R2_R0).


See the file mre/data/family.py for a description of the types of
data this implementation of MRE accepts and models. 


"""

from mats.base_mat import base_mat
from pylab import newaxis, randn, dot, exp, log, array

def fnorm(x):
    """
    if X is of dimension N * d * d, return an array Y such that
    Y[i] = frobenius_norm(X[i, :, :]).
    """
    return (x**2).sum(-1).sum(-1)

def sort_ind(x):
    """ 
    sort_ind(x):
    Sort the array x, but return the sorted indicies rather than the
    sorted values, so for example sort_ind([30, 20, 10]) returns [2, 1, 0], and 
    sort_ind([10, 20, 30]) returns [0, 1, 2].
    """
    i = range(len(x))
    c = [(u, v) for u, v in zip(x, i)]
    c.sort()
    return array(c)[:,1]

def the_smallest(x, n):
    """
def the_smallest(x, n):
    Return the indicies (ie positions) of the n smallest values in x. 
    This is useful for nearest neighbor computations.
    """
    x1=sort_ind(x)
    return x1[:n] 


## The two functions below implement MRE's gradient without being a part
## of the MRE object. They exist because of the fairly large variety
## of costs MRE optimizes; specifically, MRE minimizes several functions
## (on basic relations, on meta_obj_rel relations and on meta_rel_rel
## relations (see mre/data/family.py for a description of what these mean)).
## It's clearly better to abstract the "MRE-ness" of this whole
## thing! Hence these functions exist! 


def generic_grad(mat_a, mat_r, mat_b,
                 data, inds, alpha, 
                 grad_a, grad_r, grad_b):
        """
    def generic_grad(mat_a, mat_r, mat_b,
                     data, inds, alpha, 
                     grad_a, grad_r, grad_b):

        This is the generic MRE gradient function.

        data is an (k*3) array, where each row of data is a triple 
        (a, r, b).

        ****

        The types of a, r, and b are fixed. For example, a could be an
        object (of type O), r could be a meta relation of type (R1),
        and b could be a basic relation (thus of type R0).

        ****************

        The array mat_a contains all the matrices of the same type
        (which in the above example would be O); similarly for mat_b,
        mat_c.

        ================

        The parameter inds is a boolean vector of length(data), and
        inds[i]==True if we wish to include fact data[i] in our training. 

        We do so to deal with training/test splits.

        ================

        0<=alpha<=1 is a scalar. If alpha == 0, then we use the fully
        discriminative training that is the best when the type of r is
        R1 or R2 (ie we deal with meta relations of some kind);
        If alpha == 1 then we use the probabilistic interpretation and
        thus a partition function. This should be used when the type
        of r is R0 (ie we work with basic relations and plain objects).

        ================

        grad_x is an array of dimensions mat_x; the gradients with
        respect to mat_x are added to grad_x (and since it's an
        object, and thus a reference, it is reflected in the caller's
        object). 
        
        
        """


        for (a, r, b), use_datapoint in zip(data, inds):
            # Make sure that the data point falls into the training set.
            if not use_datapoint: 
                continue

            # load the matrix representations of the current triple
            A = mat_a[a]   
            R = mat_r[r]
            B = mat_b[b]
            
            # and get their product.
            RA = dot(R, A)
            
            # note: mat_b is an array of matrices, and fnorm, when
            # applied to an array of matrices (i.e., a 3D array)
            # produces an array of their frobenius norms.

            # unnormalized probabilities are the probabilities of
            # picking an object among those in mat_b. Note that there
            # are as many entries in unnorm_prob as there are matrices
            # in mat_b.
            unnorm_prob = exp(-fnorm(RA - mat_b)/2) 

            # these are the same unnorm_prob, normalized.
            prob = (unnorm_prob / sum(unnorm_prob))[:, newaxis, newaxis]

            # the average matrix in mat_b weighted by prob 
            # (Expected B under prob--or the expected matrix chosen by
            # RA).
            EB = (mat_b * prob).sum(0)

            aEB = alpha * EB

            # The global influence of the partition function on all of
            # mat_b: push the matrices away from RA (but only if
            # they're already close)
            grad_b += alpha * prob * (RA - mat_b)

            # and the influences on the individual matrices that take
            # part in a given relation. 
            grad_r[r] += ( dot(dot(R, A), A.T) * (1 - alpha) + 
                           dot(aEB - B, A.T) )

            grad_a[a] += ( dot(R.T, dot(R, A)) * (1 - alpha)   +
                           dot(R.T, aEB - B ) )

            grad_b[b] += B - RA


def generic_cost(mat_a, mat_r, mat_b, data, inds, alpha):
        """
        Compute the loss experienced by MRE (given all the parameters
        as inputs).

        The semantics of mat_a, mat_r, mat_b, data, and alpha are the
        same as those in generic_grad.
        """
        
        loss = 0.

        for (a, r, b), use_datapoint in zip(data, inds):
            # Again, make sure that this point falls into the training set.
            # if not, continue.
            if not use_datapoint:
                continue 


            # get the matrices
            A = mat_a[a]
            R = mat_r[r]
            B = mat_b[b]

            RA = dot(R, A)

            # and compute the log prob on this example.
            loss += (fnorm(RA - B)/2 + 
                     alpha * log(sum(exp(-fnorm(RA - mat_b)/2))))

            # as befoer, note that the log(sum(exp...)) is taken only
            # over the matrices in mat_b.
        return loss


def generic_nearest_neighbors(mat_a, mat_r, mat_b, data, inds,
                              num_neighbors):
        """
        For a given set of matrices and datasets, compute the number
        of errors experienced on this dataset.

        The first 5 parameters are as in generic_grad.
        
        The possible answers to the queries are -only- the matrices in mat_b.

        Matrices not in mat_b are not considered as possible answers.

        

        num_neighbors:

        When we do the testing, we pick a triple (a, r, b) in data,
        and ask the system: "list those who fit in the query (a, r, ?)".
        In this case, the system must know how many possible answers
        there are in reality. If there is only one possible answer
        (which can be checked by going over all the true relations),
        then the system looks at the nearest neighbor and makes sure
        that this  neighbor is r (if not, an error is added). However,
        if there are, eg 3 possible answers, then the system looks at
        the 3 nearest neighbors and makes sure that r is among them
        (if not, an error is added).


        Note that num_neighbors cannot be computed from data by this
        function becasue data could be training data, which is only a
        subset of all the true relations.

        This, in turn, would cause 
        """
        
        errors = 0
        num_cases = 0
        for (a, r, b), num_answers, use in zip(data, num_neighbors, inds):
            if not use:
                continue 

            num_cases += 1

            # num_answers is the number of answers to the query (a, r, ?)
            # if use: then this point is in the current train/test/? set.

            A = mat_a[a]
            R = mat_r[r]
            B = mat_b[b]

            RA = dot(R, A)

            dists_to_RA = fnorm(RA - mat_b) # note: we compare distances to
                                            # matrices in mat_b only!

            if b not in the_smallest(dists_to_RA, 
                                     num_answers):
                errors += 1
    
        return errors, num_cases



from pylab import zeros
def compute_num_neighbors(data):
    """
    The data consits of triples of the form (a, r, b);
    We return an array num_answers_arr of the same length, so that
    if data[i]==(a, r, b), and there are k answers to the query 
    (a, r, ?) (if we look at the array data), then num_answers_arr[i]=k.

    """
    num_answers_to_query = dict()
    for a, r, b in data:
        if (a, r) not in num_answers_to_query:
            num_answers_to_query[a, r] = 0

        num_answers_to_query[a, r] += 1

    num_answers_arr = zeros(len(data))

    for i, (a, r, b) in enumerate(data):
        num_answers_arr[i] = num_answers_to_query[a, r]

    return num_answers_arr
    


class Mre(base_mat):
    """
    This class implements MRE by calling the above functinos with
    appropriate parameters.  

    It inherits from base_mat, so it must implement the function
    soft_copy (see its doc string in base_mat). 
    The point is this: if all its parameters are stored 
    in self.w (which is a list), then a very large number of
    convenient functions will immediately become available. 
    For example, it'll be easy to convert an Mre object from and to a
    flat (1-D) vector, and it'll be possible to easily add Mre objects
    and multiply them by scalars. 
    """

    def __init__(self, 
                 dim,

                 num_O, 
                 num_R0,
                 num_R1,
                 num_R2,

                 O_R0_O,
                 O_R1_R0,
                 R0_R2_R0,

                 train_O_R0_O_inds,
                 train_O_R1_R0_inds,
                 train_R0_R2_R0_inds

                 ):
        """
        def __init__(...):
        dim: the dimensionality of the data.

        num_X : the number of entities of type X.
        For example, num_R0 is the number of the relations of type R0,
        which are the basic relations.

        O_R0_O: all known facts of the form (object, basic relation,
        object), including train and test relations.

        train_O_R2_O_inds: a boolean array, where 
        train_O_R2_O_inds[i] == True iff O_R0_O[i] is being trained
        on. 

        Thus len(train_O_R2_O_inds)==len(O_R0_O).

        ***
        Similarly for O_R1_R0 and R0_R2_R0, and their train_*_inds.
        """


        self.dim = dim

        self.O_R0_O = O_R0_O
        self.O_R1_R0 = O_R1_R0
        self.R0_R2_R0 = R0_R2_R0

        # get the training indicies
        self.train_O_R0_O_inds = train_O_R0_O_inds
        self.train_O_R1_R0_inds = train_O_R1_R0_inds
        self.train_R0_R2_R0_inds = train_R0_R2_R0_inds

        # and the test indicies
        self.test_O_R0_O_inds = ~train_O_R0_O_inds
        self.test_O_R1_R0_inds = ~train_O_R1_R0_inds
        self.test_R0_R2_R0_inds = ~train_R0_R2_R0_inds
        
        # 
        # 
        # When testing nearest neighbor performance, we must know how
        # many anwsers does a query have. Specifically, if (a, r, b)
        # is in the training set, how many answers does (a, r, ?) have?
        # To answer this query, we must have access to -all- the 
        # relations (both train and test). Once we know it, we can do
        # sensible nearset neighobr computations that tell us the
        # number of errors.
        self.num_neighbors_O_R0_O = compute_num_neighbors(self.O_R0_O)



        self.num_O = num_O
        self.num_R0 = num_R0
        self.num_R1 = num_R1
        self.num_R2 = num_R2

        self.w = [randn(num_O, dim, dim),
                  randn(num_R0, dim, dim),
                  randn(num_R1, dim, dim),
                  randn(num_R2, dim, dim)]


    def soft_copy(self):
        """
        soft_copy is a usual routine that makes a copy of the current
        object, but not of its parameters: it only returns reference
        to the existing parameters. This is necessary for the base_mat
        object to work correctly and properly. 
        """

        A = Mre( self.dim,

                 self.num_O,
                 self.num_R0,
                 self.num_R1,
                 self.num_R2,

                 self.O_R0_O,
                 self.O_R1_R0,
                 self.R0_R2_R0,

                 self.train_O_R0_O_inds,
                 self.train_O_R1_R0_inds,
                 self.train_R0_R2_R0_inds

                 )
                 

        # the parameters;
        A.w = [x for x in self.w]

        return A


    def grad(self):
        """
        Compute the gradient with respect to everything on the -train- data.
        """
        
        mat_O, mat_R0, mat_R1, mat_R2 = self

        GRAD = 0 * self

        d_O, d_R0, d_R1, d_R2 = GRAD

        generic_grad(mat_O, mat_R0, mat_O,
                     self.O_R0_O, self.train_O_R0_O_inds,
                     1., #<--alpha
                     d_O, d_R0, d_O)

        generic_grad(mat_O, mat_R1, mat_R0,
                     self.O_R1_R0, self.train_O_R1_R0_inds,
                     0., #<--alpha
                     d_O, d_R1, d_R0)

        generic_grad(mat_R0, mat_R2, mat_R0,
                     self.R0_R2_R0, self.train_R0_R2_R0_inds,
                     0., #<--alpha
                     d_R0, d_R2, d_R0)

        return GRAD


    def cost(self):
        """
        Compute the total cost on the -train- data.
        """
        
        mat_O, mat_R0, mat_R1, mat_R2 = self

        loss = (generic_cost(mat_O, mat_R0, mat_O,
                             self.O_R0_O, self.train_O_R0_O_inds,
                             1.) #<--alpha
                +

                generic_cost(mat_O, mat_R1, mat_R0,
                             self.O_R1_R0, self.train_O_R1_R0_inds,
                             0.) #<--alpha

                + 
                generic_cost(mat_R0, mat_R2, mat_R0,
                             self.R0_R2_R0, self.train_R0_R2_R0_inds,
                             0.) #<--alpha
                )
        return loss


    def nearest_neighbor(self):

        mat_O, mat_R0, mat_R1, mat_R2 = self

        train_errors = generic_nearest_neighbors(mat_O, mat_R0, mat_O, 
                                                 self.O_R0_O, self.train_O_R0_O_inds,
                                                 self.num_neighbors_O_R0_O)
        
        test_errors =  generic_nearest_neighbors(mat_O, mat_R0, mat_O, 
                                                 self.O_R0_O, self.test_O_R0_O_inds,
                                                 self.num_neighbors_O_R0_O)
        return train_errors, test_errors





    
